'''Common tools for interactive utilities such as repl or backup'''

from __future__ import print_function

import copy, getpass, os, re, socket, string, sys

from . import net
r = net.Connection._r

# This file contains common functions used by the import/export/dump/restore scripts

portRegex = re.compile(r'^\s*(?P<hostname>[\w\.-]+)(:(?P<port>\d+))?\s*$')
def parse_connect_option(connect):
    hostname = os.environ.get('RETHINKDB_HOSTNAME', 'localhost')
    port = os.environ.get('RETHINKDB_DRIVER_PORT')
    try:
        port = int(port)
    except (TypeError, ValueError): # nonsense or None
        port = net.DEFAULT_PORT
    
    if connect:
        res = portRegex.match(connect)
        if not res:
            raise ValueError("Error: Invalid 'host:port' format: %s" % connect)
        
        if res.group('hostname'):
            hostname = res.group('hostname')
        if res.group('port'):
            port = int(res.group('port'))
    
    return (hostname, port)

def parse_db_table(item):
    if not all(c in string.ascii_letters + string.digits + "._" for c in item):
        raise RuntimeError("Error: Invalid 'db' or 'db.table' name: %s" % item)
    db_table = item.split(".")
    if len(db_table) == 1:
        return (db_table[0], None)
    elif len(db_table) == 2:
        return tuple(db_table)
    else:
        raise RuntimeError("Error: Invalid 'db' or 'db.table' format: %s" % item)

def parse_db_table_options(db_table_options):
    res = []
    for item in db_table_options:
        res.append(parse_db_table(item))
    return res

def ssl_option(str):
    if str == "":
        return dict()
    else:
        return {"ca_certs": str}

def get_password(interactive, filename):
    password = ""
    if filename is not None:
        password_file = open(filename)
        password = password_file.read().rstrip('\n')
        password_file.close()
    elif interactive:
        password = getpass.getpass("Password for `admin`: ")
    return password

# This function is used to wrap rethinkdb calls to recover from connection errors
# The first argument to the function is an output parameter indicating if progress
# has been made since the last call.  This is passed as an array so it works as an
# output parameter. The first item in the array is compared each attempt to check
# if progress has been made.
# Using this wrapper, the given function will be called until 5 connection errors
# occur in a row with no progress being made.  Care should be taken that the given
# function will terminate as long as the progress parameter is changed.
def rdb_call_wrapper(conn_fn, context, fn, *args, **kwargs):
    i = 0
    max_attempts = 5
    progress = [None]
    while True:
        last_progress = copy.deepcopy(progress[0])
        try:
            conn = conn_fn()
            return fn(progress, conn, *args, **kwargs)
        except socket.error as ex:
            i = i + 1 if progress[0] == last_progress else 0
            if i == max_attempts:
                raise RuntimeError("Connection error during '%s': %s" % (context, ex.message))
        except (r.ReqlError, r.ReqlDriverError) as ex:
            raise RuntimeError("ReQL error during '%s': %s" % (context, ex.message))

def print_progress(ratio):
    total_width = 40
    done_width = int(ratio * total_width)
    undone_width = total_width - done_width
    print("\r[%s%s] %3d%%" % ("=" * done_width, " " * undone_width, int(100 * ratio)), end=' ')
    sys.stdout.flush()

def check_minimum_version(progress, conn, minimum_version):
    stringify_version = lambda v: '.'.join(map(str, v))
    parsed_version = None
    try:
        version = r.db('rethinkdb').table('server_status')[0]['process']['version'].run(conn)
        matches = re.match(r'rethinkdb (\d+)\.(\d+)\.(\d+).*', version)
        if matches == None:
            raise RuntimeError("invalid version string format")
        parsed_version = tuple(int(num) for num in matches.groups())
        if parsed_version < minimum_version:
            raise RuntimeError("incompatible version")
    except (RuntimeError, TypeError, r.ReqlRuntimeError):
        if parsed_version is None:
            message = "Error: Incompatible server version found, expected >= %s" % \
                stringify_version(minimum_version)
        else:
            message = "Error: Incompatible server version found (%s), expected >= %s" % \
                (stringify_version(parsed_version), stringify_version(minimum_version))
        raise RuntimeError(message)



# Copyright 2010-2015 RethinkDB, all rights reserved.

__all__ = [
    "ReqlCursorEmpty", "RqlCursorEmpty",
    "ReqlError", "RqlError",
        "ReqlCompileError", "RqlCompileError",
            "ReqlDriverCompileError",
            "ReqlServerCompileError",
        "ReqlRuntimeError", "RqlRuntimeError",
            "ReqlQueryLogicError",
                "ReqlNonExistenceError",
            "ReqlResourceLimitError",
            "ReqlUserError",
            "ReqlInternalError",
            "ReqlTimeoutError", "RqlTimeoutError",
            "ReqlAvailabilityError",
                "ReqlOpFailedError",
                "ReqlOpIndeterminateError",
            "ReqlPermissionError",
        "ReqlDriverError", "RqlDriverError", "RqlClientError",
            "ReqlAuthError"
]

import sys

try:
    unicode

    def convertForPrint(inputString):
        if type(inputString) == unicode:
            encoding = 'utf-8'
            if hasattr(sys.stdout, 'encoding') and sys.stdout.encoding:
                encoding = sys.stdout.encoding
            return inputString.encode(encoding or 'utf-8', 'replace')
        else:
            return str(inputString)
except NameError:
    def convertForPrint(inputString):
        return inputString

try:
    {}.iteritems
    dict_items = lambda d: d.iteritems()
except AttributeError:
    dict_items = lambda d: d.items()


class ReqlCursorEmpty(Exception):
    def __init__(self):
        super(ReqlCursorEmpty, self).__init__("Cursor is empty.")
        self.message = "Cursor is empty."

RqlCursorEmpty = ReqlCursorEmpty


class ReqlError(Exception):
    def __init__(self, message, term=None, frames=None):
        super(ReqlError, self).__init__(message)
        self.message = message
        self.frames = frames
        if term is not None and frames is not None:
            self.query_printer = QueryPrinter(term, self.frames)

    def __str__(self):
        if self.frames is None:
            return convertForPrint(self.message)
        else:
            return convertForPrint("%s in:\n%s\n%s" % (
                self.message.rstrip("."),
                self.query_printer.print_query(),
                self.query_printer.print_carrots()))
    
    def __repr__(self):
        return "<%s instance: %s >" % (self.__class__.__name__, str(self))

RqlError = ReqlError


class ReqlCompileError(ReqlError):
    pass

RqlCompileError = ReqlCompileError

class ReqlDriverCompileError(ReqlCompileError):
    pass

class ReqlServerCompileError(ReqlCompileError):
    pass


class ReqlRuntimeError(ReqlError):
    pass

RqlRuntimeError = ReqlRuntimeError


class ReqlQueryLogicError(ReqlRuntimeError):
    pass


class ReqlNonExistenceError(ReqlQueryLogicError):
    pass


class ReqlResourceLimitError(ReqlRuntimeError):
    pass


class ReqlUserError(ReqlRuntimeError):
    pass


class ReqlInternalError(ReqlRuntimeError):
    pass


class ReqlAvailabilityError(ReqlRuntimeError):
    pass


class ReqlOpFailedError(ReqlAvailabilityError):
    pass


class ReqlOpIndeterminateError(ReqlAvailabilityError):
    pass


class ReqlPermissionError(ReqlRuntimeError):
    pass


class ReqlDriverError(ReqlError):
    pass

RqlClientError = ReqlDriverError
RqlDriverError = ReqlDriverError

class ReqlAuthError(ReqlDriverError):
    def __init__(self, msg, host=None, port=None):
        if host is None or port is None:
            super(ReqlDriverError, self).__init__(msg)
        else:
            super(ReqlDriverError, self).__init__(
                "Could not connect to %s:%d: %s" %
                (host, port, msg))


try:
    class ReqlTimeoutError(ReqlDriverError, TimeoutError):
        def __init__(self, host=None, port=None):
            if host is None or port is None:
                super(ReqlDriverError, self).__init__("Operation timed out.")
            else:
                super(ReqlDriverError, self).__init__(
                    "Could not connect to %s:%d, operation timed out." % (host, port))
except NameError:
    class ReqlTimeoutError(ReqlDriverError):
        def __init__(self, host=None, port=None):
            if host is None or port is None:
                super(ReqlDriverError, self).__init__("Operation timed out.")
            else:
                super(ReqlDriverError, self).__init__(
                    "Could not connect to %s:%d, operation timed out." % (host, port))

RqlTimeoutError = ReqlTimeoutError


class QueryPrinter(object):
    def __init__(self, root, frames=[]):
        self.root = root
        self.frames = frames

    def print_query(self):
        return ''.join(self.compose_term(self.root))

    def print_carrots(self):
        return ''.join(self.compose_carrots(self.root, self.frames))

    def compose_term(self, term):
        args = [self.compose_term(a) for a in term._args]
        optargs = {}
        for k, v in dict_items(term.optargs):
            optargs[k] = self.compose_term(v)
        return term.compose(args, optargs)

    def compose_carrots(self, term, frames):
        # This term is the cause of the error
        if len(frames) == 0:
            return ['^' for i in self.compose_term(term)]

        cur_frame = frames[0]
        args = [self.compose_carrots(arg, frames[1:])
                if cur_frame == i else self.compose_term(arg)
                for i, arg in enumerate(term._args)]

        optargs = {}
        for k, v in dict_items(term.optargs):
            if cur_frame == k:
                optargs[k] = self.compose_carrots(v, frames[1:])
            else:
                optargs[k] = self.compose_term(v)

        return [' ' if i != '^' else '^' for i in term.compose(args, optargs)]


# This 'enhanced' tuple recursively iterates over it's elements allowing us to
# construct nested heirarchies that insert subsequences into tree. It's used
# to construct the query representation used by the pretty printer.
class T(object):
    # N.B Python 2.x doesn't allow keyword default arguments after *seq
    #     In Python 3.x we can rewrite this as `__init__(self, *seq, intsp=''`
    def __init__(self, *seq, **opts):
        self.seq = seq
        self.intsp = opts.pop('intsp', '')

    def __iter__(self):
        itr = iter(self.seq)
        for sub in next(itr):
            yield sub
        for token in itr:
            for sub in self.intsp:
                yield sub
            for sub in token:
                yield sub



# Copyright 2010-2016 RethinkDB, all rights reserved.

from . import version
from .ast import *
from .errors import *
from .net import *
from .query import *

__all__ = ['r', 'rethinkdb'] + errors.__all__
__version__ = version.version

try:
    import __builtin__ as builtins # Python 2
except ImportError:
    import builtins # Python 3

# The builtins here defends against re-importing something obscuring `object`.
class r(builtins.object):
    pass

for module in (net, query, ast, errors):
    for functionName in module.__all__:
        setattr(r, functionName, staticmethod(getattr(module, functionName)))
rethinkdb = r

# set the _r attribute to net.Connection
net.Connection._r = r





#!/usr/bin/env python

from __future__ import print_function

import datetime, optparse, os, shutil, sys, tarfile, tempfile, time

from . import utils_common, net
r = utils_common.r

info = ""
usage = "rethinkdb dump [-c HOST:PORT] [-p] [--password-file FILENAME] [--tls-cert FILENAME] [-f FILE] [--clients NUM] [-e (DB | DB.TABLE)]..."

def print_dump_help():
    print(''''rethinkdb dump' creates an archive of data from a RethinkDB cluster
%(usage)s
  -h [ --help ]                    print this help
  -c [ --connect ] HOST:PORT       host and client port of a rethinkdb node to connect
                                   to (defaults to localhost:%(default_port)d)
  --tls-cert FILENAME              certificate file to use for TLS encryption.
  -p [ --password ]                interactively prompt for a password required to connect.
  --password-file FILENAME         read password required to connect from file.
  -f [ --file ] FILE               file to write archive to (defaults to
                                   rethinkdb_dump_DATE_TIME.tar.gz);
                                   if FILE is -, use standard output (note that
                                   intermediate files will still be written to
                                   the --temp-dir directory)
  -e [ --export ] (DB | DB.TABLE)  limit dump to the given database or table (may
                                   be specified multiple times)
  --clients NUM_CLIENTS            number of tables to export simultaneously (default: 3)
  --temp-dir DIRECTORY             the directory to use for intermediary results
  --overwrite-file                 don't abort when file given via --file already exists
  -q [ --quiet ]                   suppress non-error messages

EXAMPLES:
rethinkdb dump -c mnemosyne:39500
  Archive all data from a cluster running on host 'mnemosyne' with a client port at 39500.

rethinkdb dump -e test -f rdb_dump.tar.gz
  Archive only the 'test' database from a local cluster into a named file.

rethinkdb dump -c hades -e test.subscribers -p
  Archive a specific table from a cluster running on host 'hades' which requires a password.
''' % {'usage':usage, 'default_port':net.DEFAULT_PORT})

def parse_options(argv):
    parser = optparse.OptionParser(add_help_option=False, usage=usage)
    parser.add_option("-c", "--connect", dest="host", metavar="host:port")
    parser.add_option("-f", "--file", dest="out_file", metavar="file", default=None)
    parser.add_option("-e", "--export", dest="tables", metavar="(db | db.table)", default=[], action="append")

    parser.add_option("--tls-cert", dest="tls_cert", metavar="TLS_CERT", default="")

    parser.add_option("--temp-dir", dest="temp_dir", metavar="directory", default=None)
    parser.add_option("--overwrite-file", dest="overwrite_file", default=False, action="store_true")
    parser.add_option("--clients", dest="clients", metavar="NUM", default=3, type="int")
    parser.add_option("-q", "--quiet", dest="quiet", default=False, action="store_true")
    parser.add_option("--debug", dest="debug", default=False, action="store_true")
    parser.add_option("-h", "--help", dest="help", default=False, action="store_true")
    parser.add_option("-p", "--password", dest="password", default=False, action="store_true")
    parser.add_option("--password-file", dest="password_file", default=None)
    
    options, args = parser.parse_args(argv)

    # Check validity of arguments
    if len(args) != 0:
        raise RuntimeError("Error: No positional arguments supported. Unrecognized option(s): %s" % args)

    if options.help:
        print_dump_help()
        exit(0)

    res = {}

    # Verify valid host:port --connect option
    (res["host"], res["port"]) = utils_common.parse_connect_option(options.host)

    res["tls_cert"] = options.tls_cert

    # Verify valid output file
    if sys.platform.startswith('win32') or sys.platform.startswith('cygwin'):
        res["temp_filename"] = "rethinkdb_dump_%s" % datetime.datetime.today().strftime("%Y-%m-%dT%H-%M-%S")
    else:
        res["temp_filename"] = "rethinkdb_dump_%s" % datetime.datetime.today().strftime("%Y-%m-%dT%H:%M:%S")

    if options.out_file == "-":
        res["out_file"] = sys.stdout
    else:
        # The output file is a real file in the file system
        if options.out_file is None:
            res["out_file"] = os.path.abspath("./" + res["temp_filename"] + ".tar.gz")
        else:
            res["out_file"] = os.path.abspath(options.out_file)

        if os.path.exists(res["out_file"]) and not options.overwrite_file:
            raise RuntimeError("Error: Output file already exists: %s" % res["out_file"])

    # Verify valid client count
    if options.clients < 1:
        raise RuntimeError("Error: invalid number of clients (%d), must be greater than zero" % options.clients)
    res["clients"] = options.clients

    # Make sure the temporary directory exists and is accessible
    res["temp_dir"] = options.temp_dir

    if res["temp_dir"] is not None:
        if not os.path.isdir(res["temp_dir"]):
            raise RuntimeError("Error: Temporary directory doesn't exist or is not a directory: %s" % res["temp_dir"])
        if not os.access(res["temp_dir"], os.W_OK):
            raise RuntimeError("Error: Temporary directory inaccessible: %s" % res["temp_dir"])

    res["tables"] = options.tables
    res["quiet"] = True if res["out_file"] is sys.stdout else options.quiet
    res["debug"] = options.debug
    res["password"] = options.password
    res["password-file"] = options.password_file
    return res

def do_export(temp_dir, options):
    from . import _export
    
    if not options["quiet"]:
        print("Exporting to directory...")
    export_args = [
        "--connect", "%s:%s" % (options["host"], options["port"]),
        "--directory", os.path.join(temp_dir, options["temp_filename"]),
        "--clients", str(options["clients"]),
        "--tls-cert", options["tls_cert"]
    ]
    if options["password"]:
        export_args.append("--password")
    if options["password-file"]:
        export_args.extend(["--password-file", options["password-file"]])
    for table in options["tables"]:
        export_args.extend(["--export", table])
    if options["debug"]:
        export_args.extend(["--debug"])
    if options["quiet"]:
        export_args.extend(["--quiet"])
    
    if _export.main(export_args) != 0:
        raise RuntimeError("Error: export failed")
    # 'Done' message will be printed by the export script (unless options["quiet"])

def do_zip(temp_dir, options):
    if not options["quiet"]:
        print("Zipping export directory...")
    start_time = time.time()

    # Below,` tarfile.open()` forces us to set either `name` or `fileobj`,
    # depending on whether the output is a real file or an open file object.
    try:
        archive = None
        if hasattr(options["out_file"], 'read'):
            archive = tarfile.open(fileobj=options["out_file"], mode="w:gz")
        else:
            archive = tarfile.open(name=options["out_file"], mode="w:gz")
        
        data_dir = os.path.join(temp_dir, options["temp_filename"])
        for curr, subdirs, files in os.walk(data_dir):
            for data_file in files:
                fullPath = os.path.join(data_dir, curr, data_file)
                archivePath = os.path.relpath(fullPath, temp_dir)
                archive.add(fullPath, arcname=archivePath)
                os.unlink(fullPath)
    finally:
        archive.close()

    if not options["quiet"]:
        print("  Done (%d seconds)" % (time.time() - start_time))

def run_rethinkdb_export(options):
    # Create a temporary directory to store the intermediary results
    temp_dir = tempfile.mkdtemp(dir=options["temp_dir"])
    res = -1

    if not options["quiet"]:
        # Print a warning about the capabilities of dump, so no one is confused (hopefully)
        print("NOTE: 'rethinkdb-dump' saves data and secondary indexes, but does *not* save")
        print(" cluster metadata.  You will need to recreate your cluster setup yourself after ")
        print(" you run 'rethinkdb-restore'.")

    try:
        do_export(temp_dir, options)
        do_zip(temp_dir, options)
    except KeyboardInterrupt:
        time.sleep(0.2)
        raise RuntimeError("Interrupted")
    finally:
        shutil.rmtree(temp_dir)

def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    try:
        options = parse_options(argv)
    except RuntimeError as ex:
        print("Usage: %s" % usage, file=sys.stderr)
        print(ex, file=sys.stderr)
        return 1

    try:
        start_time = time.time()
        run_rethinkdb_export(options)
    except RuntimeError as ex:
        print(ex, file=sys.stderr)
        return 1
    return 0

if __name__ == "__main__":
    exit(main())



#!/usr/bin/env python

'''Dispatcher for interactive functions such as repl and backup'''

import os, sys, traceback
from . import utils_common, net

def startInterpreter(prog, argv):
    import code, readline, optparse
    
    connectOptions = {}
    replVariables = {'r':utils_common.r, 'rethinkdb':utils_common.r}
    banner = 'The RethinkDB driver has been imported as `r`.'
    
    # -- get host/port setup
    
    # - parse command line
    parser = optparse.OptionParser(prog=prog)
    parser.add_option("-c", "--connect",     dest="connect",     metavar="HOST:PORT", help="host and driver port of a rethinkdb server")
    parser.add_option("-d", "--driver-port", dest="driver_port", metavar="PORT",      help="driver port of a rethinkdb server", type="int")
    parser.add_option("-n", "--host-name",   dest="hostname",    metavar="HOST",      help="host and driver port of a rethinkdb server")
    
    parser.add_option("-u", "--user",        dest="user",        metavar="USERNAME",  help="user name to connect as")
    parser.add_option("-p", "--password",    dest="password",    action="store_true", help="interactively prompt for a password  to connect")
    parser.add_option("--password-file",     dest="password",    metavar="FILENAME",  help="read password required to connect from file")
    parser.add_option("--tls-cert",          dest="tls_cert",    metavar="TLS_CERT",  help="certificate file to use for TLS encryption")
    parser.add_option("--debug",             dest="debug",       default=False,       help=optparse.SUPPRESS_HELP, action="store_true")
    
    options, _ = parser.parse_args(argv)
    
    try:
        hostname, port = utils_common.parse_connect_option(options.connect)
        connectOptions['host'] = hostname
        connectOptions['port'] = port
    except ValueError as e:
        parser.error('Error: %s' % str(e))
    
    if options.hostname:
        connectOptions['host'] = options.hostname
    
    if options.driver_port:
        connectOptions['port'] = options.driver_port
    
    if options.user:
        connectOptions['user'] = options.user
    
    if options.password is True:
        connectOptions['password'] = options.password
    elif options.password:
        try:
            connectOptions['password'] = open(options.password).read().strip()
        except IOError:
            parser.error('Error: bad value for --password-file: %s' % options.password)
    
    ssl_options = None
    if options.tls_cert:
        connectOptions['ssl'] = {'ca_certs':options.tls_cert}
    
    # -- open connection
    
    try:
        replVariables['conn'] = utils_common.r.connect(**connectOptions)
        replVariables['conn'].repl()
        banner += '''
    A connection to %s:%d has been established as `conn`
    and can be used by calling `run()` on a query without any arguments.''' % (hostname, port)
    except utils_common.r.ReqlDriverError as e:
        banner += '\nWarning: %s' % str(e)
        if options.debug:
            banner += '\n' + traceback.format_exc()
    
    # -- start interpreter
    
    code.interact(banner=banner + '\n==========', local=replVariables)
    
if __name__ == '__main__':
    if __package__ is None:
        __package__ = 'rethinkdb'
    
    # -- figure out which mode we are in
    modes = ['dump', 'export', 'import', 'index_rebuild', 'repl', 'restore']
    
    if len(sys.argv) < 2 or sys.argv[1] not in modes:
        sys.exit('ERROR: Must be called with one of the following verbs: %s' % ', '.join(modes))
    
    verb = sys.argv[1]
    prog = 'python -m rethinkdb'
    if sys.version_info < (2, 7) or (sys.version_info >= (3, 0) and sys.version_info < (3, 2)):
        prog += '.__main__' # Python versions 2.6, 3.0, and 3.1 do not support running packages
    prog += ' ' + verb
    argv = sys.argv[2:] # a bit of a hack, but it works well where we need it
    
    if verb == 'dump':
        from . import _dump
        exit(_dump.main(argv))
    elif verb == 'export':
        from . import _export
        exit(_export.main(argv))
    elif verb == 'import':
        from . import _import
        exit(_import.main(argv))
    elif verb == 'index_rebuild':
        from . import _index_rebuild
        exit(_index_rebuild.main(argv))
    elif verb == 'repl':
        startInterpreter(prog, argv)
    elif verb == 'restore':
        from . import _restore
        exit(_restore.main(argv))



# Copyright 2010-2015 RethinkDB, all rights reserved.

import collections
import errno
import imp
import numbers
import os
import socket
import ssl
import struct
import time

from . import ql2_pb2 as p

__all__ = ['connect', 'set_loop_type', 'Connection', 'Cursor', 'DEFAULT_PORT']

DEFAULT_PORT = 28015

pErrorType = p.Response.ErrorType
pResponse = p.Response.ResponseType
pQuery = p.Query.QueryType

from .ast import DB, Repl, ReQLDecoder, ReQLEncoder, expr
from .errors import *
from .handshake import *

try:
    from ssl import match_hostname, CertificateError
except ImportError:
    from .backports.ssl_match_hostname import match_hostname, CertificateError

try:
    xrange
except NameError:
    xrange = range

try:
    {}.iteritems
    dict_items = lambda d: d.iteritems()
except AttributeError:
    dict_items = lambda d: d.items()

def maybe_profile(value, res):
    if res.profile is not None:
        return {'value': value, 'profile': res.profile}
    return value

class Query(object):
    def __init__(self, type, token, term, global_optargs):
        self.type = type
        self.token = token
        self.term = term
        self.global_optargs = global_optargs

        global_optargs = global_optargs or { }
        self._json_encoder = global_optargs.pop('json_encoder', None)
        self._json_decoder = global_optargs.pop('json_decoder', None)

    def serialize(self, reql_encoder=ReQLEncoder()):
        message = [self.type]
        if self.term is not None:
            message.append(self.term)
        if self.global_optargs is not None:
            message.append(expr(self.global_optargs))
        query_str = reql_encoder.encode(message).encode('utf-8')
        query_header = struct.pack('<QL', self.token, len(query_str))
        return query_header + query_str


class Response(object):
    def __init__(self, token, json_str, reql_decoder=ReQLDecoder()):
        try:
            json_str = json_str.decode('utf-8')
        except AttributeError:
            pass               # Python3 str objects are already utf-8
        self.token = token
        full_response = reql_decoder.decode(json_str)
        self.type = full_response["t"]
        self.data = full_response["r"]
        self.backtrace = full_response.get("b", None)
        self.profile = full_response.get("p", None)
        self.error_type = full_response.get("e", None)

    def make_error(self, query):
        if self.type == pResponse.CLIENT_ERROR:
            return ReqlDriverError(self.data[0], query.term, self.backtrace)
        elif self.type == pResponse.COMPILE_ERROR:
            return ReqlServerCompileError(self.data[0], query.term, self.backtrace)
        elif self.type == pResponse.RUNTIME_ERROR:
            return {
                pErrorType.INTERNAL: ReqlInternalError,
                pErrorType.RESOURCE_LIMIT: ReqlResourceLimitError,
                pErrorType.QUERY_LOGIC: ReqlQueryLogicError,
                pErrorType.NON_EXISTENCE: ReqlNonExistenceError,
                pErrorType.OP_FAILED: ReqlOpFailedError,
                pErrorType.OP_INDETERMINATE: ReqlOpIndeterminateError,
                pErrorType.USER: ReqlUserError,
                pErrorType.PERMISSION_ERROR: ReqlPermissionError
            }.get(self.error_type, ReqlRuntimeError)(
                self.data[0], query.term, self.backtrace)
        return ReqlDriverError(("Unknown Response type %d encountered" +
                                " in a response.") % self.type)


# This class encapsulates all shared behavior between cursor implementations.
# It provides iteration over the cursor using `iter`, as well as incremental
# iteration using `next`.
#
# query - the original query that resulted in the cursor, used for:
#     query.term - the term to be used for pretty-printing backtraces
#     query.token - the token to use for subsequent CONTINUE and STOP requests
#     query.global_optargs - dictate how to format results
# items - The current list of items obtained from the server, this is
#     added to in `_extend`, which is called by the ConnectionInstance when a
#     new response arrives for this cursor.
# outstanding_requests - The number of requests that are currently awaiting
#     a response from the server.  This will typically be 0 or 1 unless the
#     cursor is exhausted, but this can be higher if `close` is called.
# threshold - a CONTINUE request will be sent when the length of `items` goes
#     below this number.
# error - indicates the current state of the cursor:
#     None - there is more data available from the server and no errors have
#         occurred yet
#     Exception - an error has occurred in the cursor and should be raised
#         to the user once all results in `items` have been returned.  This
#         will be a ReqlCursorEmpty exception if the cursor completed successfully.
#
# A class that derives from this should implement the following functions:
#     def _get_next(self, timeout):
#         where `timeout` is the maximum amount of time (in seconds) to wait for the
#         next result in the cursor before raising a ReqlTimeoutError.
#     def _empty_error(self):
#         which returns the appropriate error to be raised when the cursor is empty
class Cursor(object):
    def __init__(self, conn_instance, query, first_response, items_type = collections.deque):
        self.conn = conn_instance
        self.query = query
        self.items = items_type()
        self.outstanding_requests = 0
        self.threshold = 1
        self.error = None
        self._json_decoder = self.conn._parent._get_json_decoder(self.query)

        self.conn._cursor_cache[self.query.token] = self

        self._maybe_fetch_batch()
        self._extend_internal(first_response)

    def close(self):
        if self.error is None:
            self.error = self._empty_error()
            if self.conn.is_open():
                self.outstanding_requests += 1
                self.conn._parent._stop(self)

    @staticmethod
    def _wait_to_timeout(wait):
        if isinstance(wait, bool):
            return None if wait else 0
        elif isinstance(wait, numbers.Real) and wait >= 0:
            return wait
        else:
            raise ReqlDriverError("Invalid wait timeout '%s'" % str(wait))

    def next(self, wait=True):
        return self._get_next(Cursor._wait_to_timeout(wait))

    def _extend(self, res_buf):
        self.outstanding_requests -= 1
        self._maybe_fetch_batch()

        res = Response(self.query.token, res_buf, self._json_decoder)
        self._extend_internal(res)

    def _extend_internal(self, res):
        self.threshold = len(res.data)
        if self.error is None:
            if res.type == pResponse.SUCCESS_PARTIAL:
                self.items.extend(res.data)
            elif res.type == pResponse.SUCCESS_SEQUENCE:
                self.items.extend(res.data)
                self.error = self._empty_error()
            else:
                self.error = res.make_error(self.query)

        if self.outstanding_requests == 0 and self.error is not None:
            del self.conn._cursor_cache[res.token]

    def __summary_string(self, token, end_token):
        if len(self.items) > 10:
            val_str = token.join(map(repr, [self.items[i] for i in range(0,10)]))
        else:
            val_str = token.join(map(repr, self.items))
        if len(self.items) > 10 or self.error is None:
            val_str += end_token

        if self.error is None:
            err_str = 'streaming'
        elif isinstance(self.error, ReqlCursorEmpty):
            err_str = 'done streaming'
        else:
            err_str = 'error: %s' % repr(self.error)

        return val_str, err_str

    def __str__(self):
        val_str, err_str = self.__summary_string(",\n", ", ...\n")
        return "%s (%s):\n[\n%s]" % (object.__repr__(self), err_str, val_str)

    def __repr__(self):
        val_str, err_str = self.__summary_string(", ", ", ...")
        return "<%s.%s object at %s (%s):\n [%s]>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self)),
            err_str,
            val_str
        )

    def _error(self, message):
        # Set an error and extend with a dummy response to trigger any waiters
        if self.error is None:
            self.error = ReqlRuntimeError(message, self.query.term, [])
            dummy_response = '{"t":%d,"r":[]}' % pResponse.SUCCESS_SEQUENCE
            self._extend(dummy_response)

    def _maybe_fetch_batch(self):
        if self.error is None and \
           len(self.items) < self.threshold and \
           self.outstanding_requests == 0:
            self.outstanding_requests += 1
            self.conn._parent._continue(self)


class DefaultCursorEmpty(ReqlCursorEmpty, StopIteration):
    def __init__(self):
        super(DefaultCursorEmpty, self).__init__()


class DefaultCursor(Cursor):
    def __iter__(self):
        return self

    def __next__(self):
        return self._get_next(None)

    def _empty_error(self):
        return DefaultCursorEmpty()

    def _get_next(self, timeout):
        deadline = None if timeout is None else time.time() + timeout
        while len(self.items) == 0:
            self._maybe_fetch_batch()
            if self.error is not None:
                raise self.error
            self.conn._read_response(self.query, deadline)
        return self.items.popleft()


class SocketWrapper(object):
    def __init__(self, parent, timeout):
        self.host = parent._parent.host
        self.port = parent._parent.port
        self._read_buffer = None
        self._socket = None
        self.ssl = parent._parent.ssl

        deadline = time.time() + timeout

        try:
            self._socket = socket.create_connection((self.host, self.port), timeout)
            self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

            if len(self.ssl) > 0:
                try:
                    if hasattr(ssl, 'SSLContext'): # Python2.7 and 3.2+, or backports.ssl
                        ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                        if hasattr(ssl_context, "options"):
                            ssl_context.options |= getattr(ssl, "OP_NO_SSLv2", 0)
                            ssl_context.options |= getattr(ssl, "OP_NO_SSLv3", 0)
                        ssl_context.verify_mode = ssl.CERT_REQUIRED
                        ssl_context.check_hostname = True # redundant with match_hostname
                        ssl_context.load_verify_locations(self.ssl["ca_certs"])
                        self._socket = ssl_context.wrap_socket(self._socket, server_hostname=self.host)
                    else: # this does not disable SSLv2 or SSLv3
                        self._socket = ssl.wrap_socket(
                            self._socket, cert_reqs=ssl.CERT_REQUIRED, ssl_version=ssl.PROTOCOL_SSLv23,
                            ca_certs=self.ssl["ca_certs"])
                except IOError as err:
                    self._socket.close()
                    
                    if 'EOF occurred in violation of protocol' in str(err) or 'sslv3 alert handshake failure' in str(err):
                        # probably on an older version of OpenSSL
                        raise ReqlDriverError("SSL handshake failed, likely because Python is linked against an old version of OpenSSL that does not support either TLSv1.2 or any of the allowed ciphers. This can be worked around by lowering the security setting on the server with the options `--tls-min-protocol TLSv1 --tls-ciphers EECDH+AESGCM:EDH+AESGCM:AES256+EECDH:AES256+EDH:AES256-SHA` (see server log for more information): %s" % str(err))
                    else:
                        raise ReqlDriverError("SSL handshake failed (see server log for more information): %s" % str(err))
                try:
                    match_hostname(self._socket.getpeercert(), hostname=self.host)
                except CertificateError:
                    self._socket.close()
                    raise

            parent._parent.handshake.reset()
            response = None
            while True:
                request = parent._parent.handshake.next_message(response)
                if request is None:
                    break
                # This may happen in the `V1_0` protocol where we send two requests as
                # an optimization, then need to read each separately
                if request is not "":
                    self.sendall(request)
                
                # The response from the server is a null-terminated string
                response = b''
                while True:
                    char = self.recvall(1, deadline)
                    if char == b'\0':
                        break
                    response += char
        except (ReqlAuthError, ReqlTimeoutError):
            self.close()
            raise
        except ReqlDriverError as ex:
            self.close()
            error = str(ex)\
                .replace('receiving from', 'during handshake with')\
                .replace('sending to', 'during handshake with')
            raise ReqlDriverError(error)
        except socket.timeout as ex:
            self.close()
            raise ReqlTimeoutError(self.host, self.port)
        except Exception as ex:
            self.close()
            raise ReqlDriverError("Could not connect to %s:%s. Error: %s" %
                                  (self.host, self.port, str(ex)))

    def is_open(self):
        return self._socket is not None

    def close(self):
        if self.is_open():
            try:
                self._socket.shutdown(socket.SHUT_RDWR)
                self._socket.close()
            except Exception:
                pass
            finally:
                self._socket = None

    def recvall(self, length, deadline):
        res = b'' if self._read_buffer is None else self._read_buffer
        timeout = None if deadline is None else max(0, deadline - time.time())
        self._socket.settimeout(timeout)
        while len(res) < length:
            while True:
                try:
                    chunk = self._socket.recv(length - len(res))
                    self._socket.settimeout(None)
                    break
                except socket.timeout:
                    self._read_buffer = res
                    self._socket.settimeout(None)
                    raise ReqlTimeoutError(self.host, self.port)
                except IOError as ex:
                    if ex.errno == errno.ECONNRESET:
                        self.close()
                        raise ReqlDriverError("Connection is closed.")
                    elif ex.errno == errno.EWOULDBLOCK:
                        # This should only happen with a timeout of 0
                        raise ReqlTimeoutError(self.host, self.port)
                    elif ex.errno != errno.EINTR:
                        self.close()
                        raise ReqlDriverError(('Connection interrupted ' +
                                              'receiving from %s:%s - %s') %
                                             (self.host, self.port, str(ex)))
                except Exception as ex:
                    self.close()
                    raise ReqlDriverError('Error receiving from %s:%s - %s' %
                                         (self.host, self.port, str(ex)))
                except:
                    self.close()
                    raise
            if len(chunk) == 0:
                self.close()
                raise ReqlDriverError("Connection is closed.")
            res += chunk
        return res

    def sendall(self, data):
        offset = 0
        while offset < len(data):
            try:
                offset += self._socket.send(data[offset:])
            except IOError as ex:
                if ex.errno == errno.ECONNRESET:
                    self.close()
                    raise ReqlDriverError("Connection is closed.")
                elif ex.errno != errno.EINTR:
                    self.close()
                    raise ReqlDriverError(('Connection interrupted ' +
                                          'sending to %s:%s - %s') %
                                         (self.host, self.port, str(ex)))
            except Exception as ex:
                self.close()
                raise ReqlDriverError('Error sending to %s:%s - %s' %
                                     (self.host, self.port, str(ex)))
            except:
                self.close()
                raise

class ConnectionInstance(object):
    def __init__(self, parent):
        self._parent = parent
        self._cursor_cache = {}
        self._header_in_progress = None
        self._socket = None
        self._closing = False

    def client_port(self):
        if self.is_open():
            return self._socket._socket.getsockname()[1]

    def client_address(self):
        if self.is_open():
            return self._socket._socket.getsockname()[0]

    def connect(self, timeout):
        self._socket = SocketWrapper(self, timeout)
        return self._parent

    def is_open(self):
        return self._socket.is_open()

    def close(self, noreply_wait=False, token=None):
        self._closing = True

        # Cursors may remove themselves when errored, so copy a list of them
        for cursor in list(self._cursor_cache.values()):
            cursor._error("Connection is closed.")
        self._cursor_cache = {}

        try:
            if noreply_wait:
                noreply = Query(pQuery.NOREPLY_WAIT, token, None, None)
                self.run_query(noreply, False)
        finally:
            self._socket.close()
            self._header_in_progress = None

    def run_query(self, query, noreply):
        self._socket.sendall(query.serialize(self._parent._get_json_encoder(query)))
        if noreply:
            return None

        # Get response
        res = self._read_response(query)

        if res.type == pResponse.SUCCESS_ATOM:
            return maybe_profile(res.data[0], res)
        elif res.type in (pResponse.SUCCESS_PARTIAL,
                          pResponse.SUCCESS_SEQUENCE):
            cursor = DefaultCursor(self, query, res)
            return maybe_profile(cursor, res)
        elif res.type == pResponse.WAIT_COMPLETE:
            return None
        elif res.type == pResponse.SERVER_INFO:
            return res.data[0]
        else:
            raise res.make_error(query)

    def _read_response(self, query, deadline=None):
        token = query.token
        # We may get an async continue result, in which case we save
        # it and read the next response
        while True:
            try:
                # The first 8 bytes give the corresponding query token
                # of this response.  The next 4 bytes give the
                # expected length of this response.
                if self._header_in_progress is None:
                    self._header_in_progress \
                        = self._socket.recvall(12, deadline)
                (res_token, res_len,) \
                    = struct.unpack("<qL", self._header_in_progress)
                res_buf = self._socket.recvall(res_len, deadline)
                self._header_in_progress = None
            except KeyboardInterrupt as ex:
                # Cancel outstanding queries by dropping this connection,
                # then create a new connection for the user's convenience.
                self._parent.reconnect(noreply_wait=False)
                raise ex

            res = None

            cursor = self._cursor_cache.get(res_token)
            if cursor is not None:
                # Construct response
                cursor._extend(res_buf)
                if res_token == token:
                    return res
            elif res_token == token:
                return Response(
                    res_token, res_buf,
                    self._parent._get_json_decoder(query))
            elif not self._closing:
                # This response is corrupted or not intended for us
                self.close()
                raise ReqlDriverError("Unexpected response received.")


class Connection(object):
    _r = None
    _json_decoder = ReQLDecoder
    _json_encoder = ReQLEncoder

    def __init__(self, conn_type, host, port, db, auth_key, user, password, timeout, ssl, _handshake_version, **kwargs):
        self.db = db

        self.host = host
        try:
            self.port = int(port)
        except ValueError:
            raise ReqlDriverError("Could not convert port %r to an integer." % port)

        self.connect_timeout = timeout

        self.ssl = ssl

        self._conn_type = conn_type
        self._child_kwargs = kwargs
        self._instance = None
        self._next_token = 0

        if 'json_encoder' in kwargs:
            self._json_encoder = kwargs.pop('json_encoder')
        if 'json_decoder' in kwargs:
            self._json_decoder = kwargs.pop('json_decoder')

        if auth_key is None and password is None:
            auth_key = password = ''
        elif auth_key is None and password is not None:
            auth_key = password
        elif auth_key is not None and password is None:
            password = auth_key
        else:
            # auth_key is not None and password is not None
            raise ReqlDriverError("`auth_key` and `password` are both set.")

        if _handshake_version == 4:
            self.handshake = HandshakeV0_4(self.host, self.port, auth_key)
        else:
            self.handshake = HandshakeV1_0(
                self._json_decoder(), self._json_encoder(), self.host, self.port, user, password)

    def client_port(self):
        if self.is_open():
            return self._instance.client_port()

    def client_address(self):
        if self.is_open():
            return self._instance.client_address()

    def reconnect(self, noreply_wait=True, timeout=None):
        if timeout is None:
            timeout = self.connect_timeout

        self.close(noreply_wait)

        self._instance = self._conn_type(self, **self._child_kwargs)
        return self._instance.connect(timeout)

    # Not thread safe. Sets this connection as global state that will be used
    # by subsequence calls to `query.run`. Useful for trying out RethinkDB in
    # a Python repl environment.
    def repl(self):
        Repl.set(self)
        return self

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close(noreply_wait=False)

    def use(self, db):
        self.db = db

    def is_open(self):
        return self._instance is not None and self._instance.is_open()

    def check_open(self):
        if self._instance is None or not self._instance.is_open():
            raise ReqlDriverError('Connection is closed.')

    def close(self, noreply_wait=True):
        if self._instance is not None:
            instance = self._instance
            noreply_wait_token = self._new_token()
            self._instance = None
            self._next_token = 0
            return instance.close(noreply_wait, noreply_wait_token)

    def noreply_wait(self):
        self.check_open()
        q = Query(pQuery.NOREPLY_WAIT, self._new_token(), None, None)
        return self._instance.run_query(q, False)

    def server(self):
        self.check_open()
        q = Query(pQuery.SERVER_INFO, self._new_token(), None, None)
        return self._instance.run_query(q, False)

    def _new_token(self):
        res = self._next_token
        self._next_token += 1
        return res

    def _start(self, term, **global_optargs):
        self.check_open()
        if 'db' in global_optargs or self.db is not None:
            global_optargs['db'] = DB(global_optargs.get('db', self.db))
        q = Query(pQuery.START, self._new_token(), term, global_optargs)
        return self._instance.run_query(q, global_optargs.get('noreply', False))

    def _continue(self, cursor):
        self.check_open()
        q = Query(pQuery.CONTINUE, cursor.query.token, None, None)
        return self._instance.run_query(q, True)

    def _stop(self, cursor):
        self.check_open()
        q = Query(pQuery.STOP, cursor.query.token, None, None)
        return self._instance.run_query(q, True)

    def _get_json_decoder(self, query):
        return (query._json_decoder or self._json_decoder)(query.global_optargs)

    def _get_json_encoder(self, query):
        return (query._json_encoder or self._json_encoder)()

class DefaultConnection(Connection):
    def __init__(self, *args, **kwargs):
        Connection.__init__(self, ConnectionInstance, *args, **kwargs)

connection_type = DefaultConnection

def connect(
        host='localhost',
        port=DEFAULT_PORT,
        db=None,
        auth_key=None,
        user='admin',
        password=None,
        timeout=20,
        ssl=dict(),
        _handshake_version=10,
        **kwargs):
    conn = connection_type(host, port, db, auth_key, user, password, timeout, ssl, _handshake_version, **kwargs)
    return conn.reconnect(timeout=timeout)

def set_loop_type(library):
    global connection_type
    import pkg_resources
    
    # find module file
    manager = pkg_resources.ResourceManager()
    libPath = '%(library)s_net/net_%(library)s.py' % {'library':library}
    if not manager.resource_exists(__name__, libPath):
        raise ValueError('Unknown loop type: %r' % library)
    
    # load the module
    modulePath = manager.resource_filename(__name__, libPath)
    moduleName = 'net_%s' % library
    moduleFile, pathName, desc = imp.find_module(moduleName, [os.path.dirname(modulePath)])
    module = imp.load_module('rethinkdb.' + moduleName, moduleFile, pathName, desc)
    
    # set the connection type
    connection_type = module.Connection
    
    # cleanup
    manager.cleanup_resources()



# Copyright 2015 RethinkDB, all rights reserved.

import time
import struct

from twisted.python import log
from twisted.internet import reactor, defer
from twisted.internet.defer import inlineCallbacks, returnValue, Deferred
from twisted.internet.defer import DeferredQueue, CancelledError
from twisted.internet.protocol import ClientFactory, Protocol
from twisted.internet.endpoints import clientFromString
from twisted.internet.error import TimeoutError

from . import ql2_pb2 as p
from .net import Query, Response, Cursor, maybe_profile
from .net import Connection as ConnectionBase
from .errors import *

__all__ = ['Connection']

pResponse = p.Response.ResponseType
pQuery = p.Query.QueryType

class DatabaseProtocol(Protocol):
    WAITING_FOR_HANDSHAKE = 0
    READY = 1

    def __init__(self, factory):
        self.factory = factory
        self.state = DatabaseProtocol.WAITING_FOR_HANDSHAKE
        self._handlers = {
            DatabaseProtocol.WAITING_FOR_HANDSHAKE: self._handleHandshake,
            DatabaseProtocol.READY: self._handleResponse
        }

        self.buf = bytes()
        self.buf_expected_length = 0
        self.buf_token = None

        self.wait_for_handshake = Deferred()

        self._open = True

    def connectionMade(self):
        # Send immediately the handshake.
        self.factory.handshake.reset()
        self.transport.write(self.factory.handshake.next_message(None))

        # Defer a timer which will callback when timed out and errback the
        # wait_for_handshake. Otherwise, it will be cancelled in
        # handleHandshake.
        self._timeout_defer = reactor.callLater(self.factory.timeout,
                self._handleHandshakeTimeout)

    def connectionLost(self, reason):
        self._open = False

    def _handleHandshakeTimeout(self):
        # If we are here, we failed to do the handshake before the timeout.
        # We close the connection and raise an ReqlTimeoutError in the
        # wait_for_handshake deferred.
        self._open = False
        self.transport.loseConnection()
        self.wait_for_handshake.errback(ReqlTimeoutError())

    def _handleHandshake(self, data):
        try:
            self.buf += data
            end_index = self.buf.find(b'\0')
            if end_index != -1:
                response = self.buf[:end_index]
                self.buf = self.buf[end_index + 1:]
                request = self.factory.handshake.next_message(response)

                if request is None:
                    # We're now ready to work with real data.
                    self.state = DatabaseProtocol.READY
                    # We cancel the scheduled timeout.
                    self._timeout_defer.cancel()
                    # We callback our wait_for_handshake.
                    self.wait_for_handshake.callback(None)
                elif request != "":
                    self.transport.write(request)
        except Exception as e:
            self.wait_for_handshake.errback(e)

    def _handleResponse(self, data):
        # If we have more than one response, we should handle all of them.
        self.buf += data
        while True:
            # 1. Read the header, until we read the length of the awaited payload.
            if self.buf_expected_length == 0:
                if len(self.buf) >= 12:
                    token, length = struct.unpack('<qL', self.buf[:12])
                    self.buf_token = token
                    self.buf_expected_length = length
                    self.buf = self.buf[12:]
                else:
                    # We quit the function, it is impossible to have read the
                    # entire payload at this point.
                    return

            # 2. Buffer the data, until the size of the data match the expected
            # length provided by the header.
            if len(self.buf) < self.buf_expected_length:
                return

            self.factory.response_handler(self.buf_token, self.buf[:self.buf_expected_length])
            self.buf = self.buf[self.buf_expected_length:]
            self.buf_token = None
            self.buf_expected_length = 0

    def dataReceived(self, data):
        try:
            if self._open:
                self._handlers[self.state](data)
        except Exception as e:
            raise ReqlDriverError('Driver failed to handle received data.'
                'Error: {exc}. Dropping the connection.'.format(exc=str(e)))
            self.transport.loseConnection()

class DatabaseProtoFactory(ClientFactory):

    protocol = DatabaseProtocol

    def __init__(self, timeout, response_handler, handshake):
        self.timeout = timeout
        self.handshake = handshake
        self.response_handler = response_handler

    def startedConnecting(self, connector):
        pass

    def buildProtocol(self, addr):
        p = DatabaseProtocol(self)
        return p

    def clientConnectionLost(self, connector, reason):
        pass

    def clientConnectionFailed(self, connector, reason):
        pass

class CursorItems(DeferredQueue):
    def __init__(self):
        super(CursorItems, self).__init__()

    def cancel_getters(self, err):
        """
        Cancel all waiters.
        """
        for waiter in self.waiting[:]:
            if not waiter.called:
                waiter.errback(err)
            self.waiting.remove(waiter)

    def extend(self, data):
        for k in data:
            self.put(k)

    def __len__(self):
        return len(self.pending)

    def __getitem__(self, index):
        return self.pending[index]

    def __iter__(self):
        return iter(self.pending)

class TwistedCursor(Cursor):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('items_type', CursorItems)
        super(TwistedCursor, self).__init__(*args, **kwargs)
        self.waiting = list()

    def _extend(self, res):
        Cursor._extend(self, res)

        if self.error is not None:
            self.items.cancel_getters(self.error)

        for d in self.waiting[:]:
            d.callback(None)
            self.waiting.remove(d)

    def _empty_error(self):
        return RqlCursorEmpty()

    @inlineCallbacks
    def fetch_next(self, wait=True):
        timeout = Cursor._wait_to_timeout(wait)
        deadline = None if timeout is None else time.time() + timeout

        def wait_canceller(d):
            d.errback(ReqlTimeoutError())

        while len(self.items) == 0 and self.error is None:
            self._maybe_fetch_batch()

            wait = Deferred(canceller=wait_canceller)
            self.waiting.append(wait)
            if deadline is not None:
                timeout = max(0, deadline - time.time())
                reactor.callLater(timeout, lambda: wait.cancel())
            yield wait

        returnValue(not self._is_empty() or self._has_error())

    def _has_error(self):
        return self.error and (not isinstance(self.error, RqlCursorEmpty))

    def _is_empty(self):
        return isinstance(self.error, RqlCursorEmpty) and len(self.items) == 0

    def _get_next(self, timeout):
        if len(self.items) == 0 and self.error is not None:
            return defer.fail(self.error)

        def raise_timeout(errback):
            if isinstance(errback.value, CancelledError):
                raise ReqlTimeoutError()
            else:
                raise errback.value

        item_defer = self.items.get()

        if timeout is not None:
            item_defer.addErrback(raise_timeout)
            timer = reactor.callLater(timeout, lambda: item_defer.cancel())

        self._maybe_fetch_batch()
        return item_defer


class ConnectionInstance(object):

    def __init__(self, parent, start_reactor=False):
        self._parent = parent
        self._closing = False
        self._connection = None
        self._user_queries = {}
        self._cursor_cache = {}

        if start_reactor:
            reactor.run()

    def client_port(self):
        if self.is_open():
            return self._connection.transport.getHost().port

    def client_address(self):
        if self.is_open():
            return self._connection.transport.getHost().host

    def _handleResponse(self, token, data):
        try:
            cursor = self._cursor_cache.get(token)
            if cursor is not None:
                cursor._extend(data)
            elif token in self._user_queries:
                query, deferred = self._user_queries[token]
                res = Response(token, data,
                               self._parent._get_json_decoder(query))
                if res.type == pResponse.SUCCESS_ATOM:
                    deferred.callback(maybe_profile(res.data[0], res))
                elif res.type in (pResponse.SUCCESS_SEQUENCE,
                                  pResponse.SUCCESS_PARTIAL):
                    cursor = TwistedCursor(self, query, res)
                    deferred.callback(maybe_profile(cursor, res))
                elif res.type == pResponse.WAIT_COMPLETE:
                    deferred.callback(None)
                elif res.type == pResponse.SERVER_INFO:
                    deferred.callback(res.data[0])
                else:
                    deferred.errback(res.make_error(query))
                del self._user_queries[token]
            elif not self._closing:
                raise ReqlDriverError("Unexpected response received.")
        except Exception as e:
            if not self._closing:
                self.close(exception=e)

    @inlineCallbacks
    def _connectTimeout(self, factory, timeout):
        try:
            # TODO: use ssl options
            # TODO: this doesn't work for literal IPv6 addresses like '::1'
            args = "tcp:%s:%d" % (self._parent.host, self._parent.port)

            if timeout is not None:
                args = args + (":timeout=%d" % timeout)

            endpoint = clientFromString(reactor, args)
            p = yield endpoint.connect(factory)
            returnValue(p)
        except TimeoutError:
            raise ReqlTimeoutError()

    @inlineCallbacks
    def connect(self, timeout):
        factory = DatabaseProtoFactory(timeout, self._handleResponse,
                self._parent.handshake)

        # We connect to the server, and send the handshake payload.
        pConnection = None
        try:
            pConnection = yield self._connectTimeout(factory, timeout)
        except Exception as e:
            raise ReqlDriverError('Could not connect to {p.host}:{p.port}. Error: {exc}'
                                  .format(p=self._parent, exc=str(e)))

        # Now, we need to wait for the handshake.
        try:
            yield pConnection.wait_for_handshake
        except ReqlAuthError as e:
            raise
        except ReqlTimeoutError as e:
            raise ReqlTimeoutError(self._parent.host, self._parent.port)
        except Exception as e:
            raise ReqlDriverError('Connection interrupted during handshake with {p.host}:{p.port}. Error: {exc}'
                                  .format(p=self._parent, exc=str(e)))

        self._connection = pConnection

        returnValue(self._parent)

    def is_open(self):
        return self._connection._open

    def close(self, noreply_wait=False, token=None, exception=None):
        d = defer.succeed(None)
        self._closing = True
        error_message = "Connection is closed"
        if exception is not None:
            error_message = "Connection is closed (reason: {exc})".format(exc=str(exception))

        for cursor in list(self._cursor_cache.values()):
            cursor._error(error_message)

        for query, deferred in iter(self._user_queries.values()):
            if not deferred.called:
                deferred.errback(fail=ReqlDriverError(error_message))

        self._user_queries = {}
        self._cursor_cache = {}

        if noreply_wait:
            noreply = Query(pQuery.NOREPLY_WAIT, token, None, None)
            d = self.run_query(noreply, False)

        def closeConnection(res):
            self._connection.transport.loseConnection()
            return res

        return d.addBoth(closeConnection)

    @inlineCallbacks
    def run_query(self, query, noreply):
        response_defer = Deferred()
        if not noreply:
            self._user_queries[query.token] = (query, response_defer)
        # Send the query
        self._connection.transport.write(query.serialize(self._parent._get_json_encoder(query)))

        if noreply:
            returnValue(None)
        else:
            res = yield response_defer
            returnValue(res)


class Connection(ConnectionBase):

    def __init__(self, *args, **kwargs):
        super(Connection, self).__init__(ConnectionInstance, *args, **kwargs)

    @inlineCallbacks
    def reconnect(self, noreply_wait=True, timeout=None):
        yield self.close(noreply_wait)
        res = yield super(Connection, self).reconnect(noreply_wait, timeout)
        returnValue(res)

    @inlineCallbacks
    def close(self, *args, **kwargs):
        res = yield super(Connection, self).close(*args, **kwargs) or None
        returnValue(res)

    @inlineCallbacks
    def noreply_wait(self, *args, **kwargs):
        res = yield super(Connection, self).noreply_wait(*args, **kwargs)
        returnValue(res)

    @inlineCallbacks
    def server(self, *args, **kwargs):
        res = yield super(Connection, self).server(*args, **kwargs)
        returnValue(res)

    @inlineCallbacks
    def _start(self, *args, **kwargs):
        res = yield super(Connection, self)._start(*args, **kwargs)
        returnValue(res)

    @inlineCallbacks
    def _continue(self, *args, **kwargs):
        res = yield super(Connection, self)._continue(*args, **kwargs)
        returnValue(res)

    @inlineCallbacks
    def _stop(self, *args, **kwargs):
        res = yield super(Connection, self)._stop(*args, **kwargs)
        returnValue(res)



import base64
import binascii
import hashlib
import hmac
import random
import struct
import sys
import threading

from . import ql2_pb2
from .errors import *

try:
    xrange
except NameError:
    xrange = range

class HandshakeV0_4(object):
    VERSION = ql2_pb2.VersionDummy.Version.V0_4
    PROTOCOL = ql2_pb2.VersionDummy.Protocol.JSON

    def __init__(self, host, port, auth_key):
        self._host = host
        self._port = port
        self._auth_key = auth_key.encode("ascii")

        self._state = 0

    def reset(self):
        self._state = 0

    def next_message(self, response):
        if self._state == 0:
            if response is not None:
                raise ReqlDriverError("Unexpected response")
            self._state = 1
            return \
                struct.pack("<2L", self.VERSION, len(self._auth_key)) + \
                self._auth_key + \
                struct.pack("<L", self.PROTOCOL)
        elif self._state == 1:
            if response is None:
                raise ReqlDriverError("Expected response")
            self._state = 2
            if response != b"SUCCESS":
                # This is an error case, we attempt to decode `response` as UTF-8 with
                # with fallbacks to get something useful
                message = None
                try:
                    message = response.decode("utf-8", errors="ignore").strip()
                except TypeError:
                    try:
                        message = response.decode("utf-8").strip()
                    except UnicodeError:
                        message = repr(response).strip()
                if message == "ERROR: Incorrect authorization key.":
                    raise ReqlAuthError("Incorrect authentication key.", self._host, self._port)
                else:
                    raise ReqlDriverError(
                        "Server dropped connection with message: \"%s\"" % (message, ))
            return None
        else:
            raise ReqlDriverError("Unexpected handshake state")

class HandshakeV1_0(object):
    VERSION = ql2_pb2.VersionDummy.Version.V1_0
    PROTOCOL = ql2_pb2.VersionDummy.Protocol.JSON

    def __init__(self, json_decoder, json_encoder, host, port, username, password):
        self._json_decoder = json_decoder
        self._json_encoder = json_encoder
        self._host = host
        self._port = port
        self._username = \
            username.encode("utf-8").replace(b"=", b"=3D").replace(b",", b"=2C")
        try:
            self._password = bytes(password, "utf-8")
        except TypeError:
            self._password = bytes(password)

        # Python 2.7.7+ and Python 3.3+
        self._compare_digest = getattr(hmac, "compare_digest", None)
        if not self._compare_digest:
            self._compare_digest = self.__compare_digest

        # Python 2.7.8+ and Python 3.4+.
        self._pbkdf2_hmac = getattr(hashlib, "pbkdf2_hmac", None)
        if not self._pbkdf2_hmac:
            self._pbkdf2_hmac = HandshakeV1_0.__pbkdf2_hmac

        self._protocol_version = 0
        self._random = random.SystemRandom()
        self._state = 0

    def reset(self):
        self._r = None
        self._client_first_message_bare = None
        self._server_signature = None
        self._state = 0

    def next_message(self, response):
        if self._state == 0:
            if response is not None:
                raise ReqlDriverError("Unexpected response")

            # Using base64 encoding for printable characters
            self._r = base64.standard_b64encode(
                bytes(bytearray(self._random.getrandbits(8) for i in xrange(18))))

            self._client_first_message_bare = b"n=" + self._username + b",r=" + self._r

            # Here we send the version as well as the initial JSON as an optimization
            self._state = 1
            return struct.pack("<L", self.VERSION) + \
                self._json_encoder.encode({
                    "protocol_version": self._protocol_version,
                    "authentication_method": "SCRAM-SHA-256",
                    "authentication":
                        (b"n,," + self._client_first_message_bare).decode("ascii")
                }).encode("utf-8") + \
                b'\0'
        elif self._state == 1:
            response = response.decode("utf-8")
            if response.startswith("ERROR"):
                raise ReqlDriverError("Received an unexpected reply.  You may be attempting to connect to a RethinkDB server that is too old for this driver.  The minimum supported server version is 2.3.0.")
            json = self._json_decoder.decode(response)
            try:
                if json["success"] == False:
                    if 10 <= json["error_code"] <= 20:
                        raise ReqlAuthError(json["error"], self._host, self._port)
                    else:
                        raise ReqlDriverError(json["error"])

                min = json["min_protocol_version"]
                max = json["max_protocol_version"]
                if not min <= self._protocol_version <= max:
                    raise ReqlDriverError(
                        "Unsupported protocol version %d, expected between %d and %d" % (
                            self._protocol_version, min, max))
            except KeyError as key_error:
                raise ReqlDriverError("Missing key: %s" % (key_error, ))

            # We've already sent the initial JSON above, and only support a single
            # protocol version at the moment thus we simply read the next response
            self._state = 2
            return ""
        elif self._state == 2:
            json = self._json_decoder.decode(response.decode("utf-8"))
            server_first_message = r = salt = i = None
            try:
                if json["success"] == False:
                    if 10 <= json["error_code"] <= 20:
                        raise ReqlAuthError(json["error"], self._host, self._port)
                    else:
                        raise ReqlDriverError(json["error"])

                server_first_message = json["authentication"].encode("ascii")
                authentication = dict(
                    x.split(b"=", 1) for x in server_first_message.split(b","))

                r = authentication[b"r"]
                if not r.startswith(self._r):
                    raise ReqlAuthError("Invalid nonce from server", self._host, self._port)
                salt = base64.standard_b64decode(authentication[b"s"])
                i = int(authentication[b"i"])
            except KeyError as key_error:
                raise ReqlDriverError("Missing key: %s" % (key_error, ))

            client_final_message_without_proof = b"c=biws,r=" + r

            # SaltedPassword := Hi(Normalize(password), salt, i)
            salted_password = self._pbkdf2_hmac("sha256", self._password, salt, i)

            # ClientKey := HMAC(SaltedPassword, "Client Key")
            client_key = hmac.new(
                salted_password, b"Client Key", hashlib.sha256).digest()

            # StoredKey := H(ClientKey)
            stored_key = hashlib.sha256(client_key).digest()

            # AuthMessage := client-first-message-bare + "," +
            #                server-first-message + "," +
            #                client-final-message-without-proof
            auth_message = b",".join((
                self._client_first_message_bare,
                server_first_message,
                client_final_message_without_proof))

            # ClientSignature := HMAC(StoredKey, AuthMessage)
            client_signature = hmac.new(
                stored_key, auth_message, hashlib.sha256).digest()

            # ClientProof := ClientKey XOR ClientSignature
            client_proof = struct.pack(
                "32B",
                *(l ^ r for l, r in zip(
                    struct.unpack("32B", client_key),
                    struct.unpack("32B", client_signature))))

            # ServerKey := HMAC(SaltedPassword, "Server Key")
            server_key = hmac.new(
                salted_password, b"Server Key", hashlib.sha256).digest()

            # ServerSignature := HMAC(ServerKey, AuthMessage)
            self._server_signature = hmac.new(
                server_key, auth_message, hashlib.sha256).digest()

            self._state = 3
            return self._json_encoder.encode({
                "authentication": (
                        client_final_message_without_proof +
                        b",p=" + base64.standard_b64encode(client_proof)
                    ).decode("ascii")
            }).encode("utf-8") + \
            b'\0'
        elif self._state == 3:
            json = self._json_decoder.decode(response.decode("utf-8"))
            v = None
            try:
                if json["success"] == False:
                    if 10 <= json["error_code"] <= 20:
                        raise ReqlAuthError(json["error"], self._host, self._port)
                    else:
                        raise ReqlDriverError(json["error"])

                authentication = dict(
                    x.split(b"=", 1)
                        for x in json["authentication"].encode("ascii").split(b","))

                v = base64.standard_b64decode(authentication[b"v"])
            except KeyError as key_error:
                raise ReqlDriverError("Missing key: %s" % (key_error, ))

            if not self._compare_digest(v, self._server_signature):
                raise ReqlAuthError("Invalid server signature", self._host, self._port)

            self._state = 4
            return None
        else:
            raise ReqlDriverError("Unexpected handshake state");

    @staticmethod
    def __compare_digest(a, b):
        if sys.version_info[0] == 3:
            def xor_bytes(a, b):
                return a ^ b
        else:
            def xor_bytes(a, b, _ord=ord):
                return _ord(a) ^ _ord(b)

        left = None
        right = b
        if len(a) == len(b):
            left = a
            result = 0
        if len(a) != len(b):
            left = b
            result = 1

        for l, r in zip(left, right):
            result |= xor_bytes(l, r)

        return result == 0

    class thread_local_cache(threading.local):
        def __init__(self):
            self.cache = {}

        def set(self, key, val):
            self.cache[key] = val

        def get(self, key):
            return self.cache.get(key)

    pbkdf2_cache = thread_local_cache()

    @staticmethod
    def __pbkdf2_hmac(hash_name, password, salt, iterations):
        assert hash_name == "sha256", hash_name

        def from_bytes(value, hexlify=binascii.hexlify, int=int):
            return int(hexlify(value), 16)

        def to_bytes(value, unhexlify=binascii.unhexlify):
            try:
                return unhexlify(bytes("%064x" % value, "ascii"))
            except TypeError:
                return unhexlify(bytes("%064x" % value))

        cache_key = (password, salt, iterations)

        cache_result = HandshakeV1_0.pbkdf2_cache.get(cache_key)

        if cache_result is not None:
            return cache_result

        mac = hmac.new(password, None, hashlib.sha256)

        def digest(msg, mac=mac):
            mac_copy = mac.copy()
            mac_copy.update(msg)
            return mac_copy.digest()

        t = digest(salt + b"\x00\x00\x00\x01")
        u = from_bytes(t)
        for c in xrange(iterations - 1):
            t = digest(t)
            u ^= from_bytes(t)

        u = to_bytes(u)
        HandshakeV1_0.pbkdf2_cache.set(cache_key, u)
        return u



#!/usr/bin/env python

from __future__ import print_function

import datetime, optparse, os, shutil, sys, tarfile, tempfile, time
from . import utils_common, net

info = "'rethinkdb restore' loads data into a RethinkDB cluster from an archive"
usage = "rethinkdb restore FILE [-c HOST:PORT] [--tls-cert FILENAME] [-p] [--password-file FILENAME] [--clients NUM] [--shards NUM_SHARDS] [--replicas NUM_REPLICAS] [--force] [-i (DB | DB.TABLE)]..."

def print_restore_help():
    print(info)
    print(usage)
    print("")
    print("  FILE                             the archive file to restore data from;")
    print("                                   if FILE is -, use standard input (note that")
    print("                                   intermediate files will still be written to")
    print("                                   the --temp-dir directory)")
    print("  -h [ --help ]                    print this help")
    print("  -c [ --connect ] HOST:PORT       host and client port of a rethinkdb node to connect")
    print("                                   to (defaults to localhost:%d)" % net.DEFAULT_PORT)
    print("  --tls-cert FILENAME              certificate file to use for TLS encryption.")
    print("  -p [ --password ]                interactively prompt for a password required to connect.")
    print("  --password-file FILENAME         read password required to connect from file.")
    print("  -i [ --import ] (DB | DB.TABLE)  limit restore to the given database or table (may")
    print("                                   be specified multiple times)")
    print("  --clients NUM_CLIENTS            the number of client connections to use (defaults")
    print("                                   to 8)")
    print("  --temp-dir DIRECTORY             the directory to use for intermediary results")
    print("  --hard-durability                use hard durability writes (slower, but less memory")
    print("                                   consumption on the server)")
    print("  --force                          import data even if a table already exists")
    print("  --no-secondary-indexes           do not create secondary indexes for the restored tables")
    print("  -q [ --quiet ]                   suppress non-error messages")
    print("")
    print("EXAMPLES:")
    print("")
    print("rethinkdb restore rdb_dump.tar.gz -c mnemosyne:39500")
    print("  Import data into a cluster running on host 'mnemosyne' with a client port at 39500 using")
    print("  the named archive file.")
    print("")
    print("rethinkdb restore rdb_dump.tar.gz -i test")
    print("  Import data into a local cluster from only the 'test' database in the named archive file.")
    print("")
    print("rethinkdb restore rdb_dump.tar.gz -i test.subscribers -c hades -p")
    print("  Import data into a cluster running on host 'hades' which requires a password from only")
    print("  a specific table from the named archive file.")
    print("")
    print("rethinkdb restore rdb_dump.tar.gz --clients 4 --force")
    print("  Import data to a local cluster from the named archive file using only 4 client connections")
    print("  and overwriting any existing rows with the same primary key.")

def parse_options(argv):
    parser = optparse.OptionParser(add_help_option=False, usage=usage)
    parser.add_option("-c", "--connect", dest="host", metavar="HOST:PORT")
    parser.add_option("-i", "--import", dest="tables", metavar="DB | DB.TABLE", default=[], action="append")

    parser.add_option("--shards", dest="shards", metavar="NUM_SHARDS", default=0, type="int")
    parser.add_option("--replicas", dest="replicas", metavar="NUM_REPLICAS", default=0, type="int")

    parser.add_option("--tls-cert", dest="tls_cert", metavar="TLS_CERT", default="")

    parser.add_option("--temp-dir", dest="temp_dir", metavar="directory", default=None)
    parser.add_option("--clients", dest="clients", metavar="NUM_CLIENTS", default=8, type="int")
    parser.add_option("--hard-durability", dest="hard", action="store_true", default=False)
    parser.add_option("--force", dest="force", action="store_true", default=False)
    parser.add_option("--no-secondary-indexes", dest="create_sindexes", action="store_false", default=True)
    parser.add_option("-q", "--quiet", dest="quiet", default=False, action="store_true")
    parser.add_option("--debug", dest="debug", default=False, action="store_true")
    parser.add_option("-h", "--help", dest="help", default=False, action="store_true")
    parser.add_option("-p", "--password", dest="password", default=False, action="store_true")
    parser.add_option("--password-file", dest="password_file", default=None)
    (options, args) = parser.parse_args(argv)

    if options.help:
        print_restore_help()
        exit(0)

    # Check validity of arguments
    if len(args) == 0:
        raise RuntimeError("Error: Archive to import not specified.  Provide an archive file from rethinkdb-dump.")
    elif len(args) != 1:
        raise RuntimeError("Error: Only one positional argument supported")

    res = {}

    # Verify valid host:port --connect option
    (res["host"], res["port"]) = utils_common.parse_connect_option(options.host)

    in_file_argument = args[0]

    if in_file_argument == "-":
        res["in_file"] = sys.stdin
    else:
        # Verify valid input file
        res["in_file"] = os.path.abspath(in_file_argument)

        if not os.path.exists(res["in_file"]):
            raise RuntimeError("Error: Archive file does not exist: %s" % res["in_file"])

    # Verify valid --import options
    res["tables"] = utils_common.parse_db_table_options(options.tables)

    # Make sure the temporary directory exists and is accessible
    res["temp_dir"] = options.temp_dir

    if res["temp_dir"] is not None:
        if not os.path.isdir(res["temp_dir"]):
            raise RuntimeError("Error: Temporary directory doesn't exist or is not a directory: %s" % res["temp_dir"])
        if not os.access(res["temp_dir"], os.W_OK):
            raise RuntimeError("Error: Temporary directory inaccessible: %s" % res["temp_dir"])

    res["clients"] = options.clients
    res["shards"] = options.shards
    res["replicas"] = options.replicas
    res["hard"] = options.hard
    res["force"] = options.force
    res["create_sindexes"] = options.create_sindexes
    res["quiet"] = options.quiet
    res["debug"] = options.debug

    res["tls_cert"] = options.tls_cert
    res["password"] = options.password
    res["password-file"] = options.password_file
    return res

def do_unzip(temp_dir, options):
    if not options["quiet"]:
        print("Unzipping archive file...")
    start_time = time.time()
    original_dir = os.getcwd()
    sub_path = None

    tables_to_export = set(options["tables"])
    members = []

    def parse_path(member):
        path = os.path.normpath(member.name)
        if member.isdir():
            return ('', '', '')
        if not member.isfile():
            raise RuntimeError("Error: Archive file contains an unexpected entry type")
        if not os.path.realpath(os.path.abspath(path)).startswith(os.getcwd()):
            raise RuntimeError("Error: Archive file contains unexpected absolute or relative path")
        path, table_file = os.path.split(path)
        path, db = os.path.split(path)
        path, base = os.path.split(path)
        return (base, db, os.path.splitext(table_file)[0])

    # If the in_file is a fileobj (e.g. stdin), stream it to a seekable file
    # first so that the code below can seek in it.
    tar_temp_file_path = None
    if options["in_file"] is sys.stdin:
        fileobj = options["in_file"]
        fd, tar_temp_file_path = tempfile.mkstemp(suffix=".tar.gz", dir=options["temp_dir"])
        # Constant memory streaming, buf == "" on EOF
        with os.fdopen(fd, "w", 65536) as f:
            buf = None
            while buf != "":
                buf = fileobj.read(65536)
                f.write(buf)
        name = tar_temp_file_path
    else:
        name = options["in_file"]

    try:
        with tarfile.open(name, "r:*") as f:
            os.chdir(temp_dir)
            try:
                for member in f:
                    base, db, table = parse_path(member)
                    if len(base) > 0:
                        if sub_path is None:
                            sub_path = base
                        elif sub_path != base:
                            raise RuntimeError("Error: Archive file has an unexpected directory structure (%s vs %s)" % (sub_path, base))

                        if len(tables_to_export) == 0 or \
                           (db, table) in tables_to_export or \
                           (db, None) in tables_to_export:
                            members.append(member)

                if sub_path is None:
                    raise RuntimeError("Error: Archive file had no files")

                f.extractall(members=members)

            finally:
                os.chdir(original_dir)
    finally:
        if tar_temp_file_path is not None:
            os.remove(tar_temp_file_path)

    if not options["quiet"]:
        print("  Done (%d seconds)" % (time.time() - start_time))
    return os.path.join(temp_dir, sub_path)

def do_import(temp_dir, options):
    from . import _import
    
    if not options["quiet"]:
        print("Importing from directory...")

    import_args = [
        "--connect", "%s:%s" % (options["host"], options["port"]),
        "--directory", temp_dir,
        "--clients", str(options["clients"]),
        "--shards", str(options["shards"]),
        "--replicas", str(options["replicas"]),
        "--tls-cert", options["tls_cert"]
    ]
    if options["password"]:
        import_args.append("--password")
    if options["password-file"]:
        import_args.extend(["--password-file", options["password-file"]])
    
    for db, table in options["tables"]:
        if table is None:
            import_args.extend(["--import", db])
        else:
            import_args.extend(["--import", "%s.%s" % (db, table)])

    if options["hard"]:
        import_args.append("--hard-durability")
    if options["force"]:
        import_args.append("--force")
    if options["debug"]:
        import_args.extend(["--debug"])
    if not options["create_sindexes"]:
        import_args.extend(["--no-secondary-indexes"])
    if options["quiet"]:
        import_args.extend(["--quiet"])
    
    res = _import.main(import_args)

    if res == 2:
        raise RuntimeError("Warning: import did not create some secondary indexes.")
    elif res != 0:
        raise RuntimeError("Error: import failed")
    # 'Done' message will be printed by the import script

def run_rethinkdb_import(options):
    # Create a temporary directory to store the extracted data
    temp_dir = tempfile.mkdtemp(dir=options["temp_dir"])
    res = -1

    try:
        sub_dir = do_unzip(temp_dir, options)
        do_import(sub_dir, options)
    except KeyboardInterrupt:
        time.sleep(0.2)
        raise RuntimeError("Interrupted")
    finally:
        shutil.rmtree(temp_dir)

def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    try:
        options = parse_options(argv)
    except RuntimeError as ex:
        print("Usage: %s" % usage, file=sys.stderr)
        print(ex, file=sys.stderr)
        return 1

    try:
        start_time = time.time()
        run_rethinkdb_import(options)
    except RuntimeError as ex:
        print(ex, file=sys.stderr)
        return 1
    return 0

if __name__ == "__main__":
    exit(main())



# Copyright 2010-2015 RethinkDB, all rights reserved.
"""
Please see the following links for more information::

- `Python API documentation <http://rethinkdb.com/api/python/>`_
- `GitHub Project <https://github.com/rethinkdb/rethinkdb>`_
- `Python Driver Source <https://github.com/rethinkdb/rethinkdb/tree/next/drivers/python>`_

"""

import os, setuptools, sys

modulePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'rethinkdb')
sys.path.insert(0, modulePath)
from version import version
sys.path.remove(modulePath)

conditionalPackages = []
if 'upload' in sys.argv: # ensure that uplodas always have everything while uploading to pypi
    conditionalPackages = ['rethinkdb.asyncio_net']
else:
    try: # only add asyncio when it is supported per #4702
        import asyncio
        conditionalPackages = ['rethinkdb.asyncio_net']
    except ImportError: pass

setuptools.setup(
    name="rethinkdb",
    zip_safe=True,
    version=version,
    description="Python driver library for the RethinkDB database server.",
    long_description=__doc__,
    url="http://rethinkdb.com",
    maintainer="RethinkDB Inc.",
    maintainer_email="bugs@rethinkdb.com",
    packages=[
        'rethinkdb',
        'rethinkdb.tornado_net',
        'rethinkdb.twisted_net',
        'rethinkdb.gevent_net',
        'rethinkdb.backports',
        'rethinkdb.backports.ssl_match_hostname'
    ] + conditionalPackages,
    package_dir={'rethinkdb':'rethinkdb'},
    package_data={ 'rethinkdb':['backports/ssl_match_hostname/*.txt'] },
    entry_points={
        'console_scripts':[
            'rethinkdb-import = rethinkdb._import:main',
            'rethinkdb-dump = rethinkdb._dump:main',
            'rethinkdb-export = rethinkdb._export:main',
            'rethinkdb-restore = rethinkdb._restore:main',
            'rethinkdb-index-rebuild = rethinkdb._index_rebuild:main'
        ]
    }
)



# Copyright 2015 RethinkDB, all rights reserved.

import errno
import struct
import gevent
import gevent.socket as socket
from gevent.event import Event, AsyncResult
from gevent.lock import Semaphore

from . import ql2_pb2 as p
from . import net
from .errors import *

__all__ = ['Connection']

pResponse = p.Response.ResponseType
pQuery = p.Query.QueryType


class GeventCursorEmpty(ReqlCursorEmpty, StopIteration):
    pass


# TODO: allow users to set sync/async?
class GeventCursor(net.Cursor):
    def __init__(self, *args, **kwargs):
        super(GeventCursor, self).__init__(*args, **kwargs)
        self.new_response = Event()

    def __iter__(self):
        return self

    def __next__(self):
        return self._get_next(None)

    def _empty_error(self):
        return GeventCursorEmpty()

    def _extend(self, res):
        super(GeventCursor, self)._extend(res)
        self.new_response.set()
        self.new_response.clear()

    def _get_next(self, timeout):
        with gevent.Timeout(timeout, RqlTimeoutError()) as timeout:
            self._maybe_fetch_batch()
            while len(self.items) == 0:
                if self.error is not None:
                    raise self.error
                self.new_response.wait()
            return self.items.popleft()


# TODO: would be nice to share this code with net.py
class SocketWrapper(net.SocketWrapper):
    def __init__(self, parent):
        self.host = parent._parent.host
        self.port = parent._parent.port
        self._read_buffer = None
        self._socket = None
        self.ssl = parent._parent.ssl

        try:
            self._socket = socket.create_connection((self.host, self.port))
            self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

            if len(self.ssl) > 0:
                try:
                    if hasattr(ssl, 'SSLContext'): # Python2.7 and 3.2+, or backports.ssl
                        ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                        if hasattr(ssl_context, "options"):
                            ssl_context.options |= getattr(ssl, "OP_NO_SSLv2", 0)
                            ssl_context.options |= getattr(ssl, "OP_NO_SSLv3", 0)
                        self.ssl_context.verify_mode = ssl.CERT_REQUIRED
                        self.ssl_context.check_hostname = True # redundant with match_hostname
                        self.ssl_context.load_verify_locations(self.ssl["ca_certs"])
                        self._socket = ssl_context.wrap_socket(self._socket, server_hostname=self.host)
                    else: # this does not disable SSLv2 or SSLv3
                        self._socket = ssl.wrap_socket(
                            self._socket, cert_reqs=ssl.CERT_REQUIRED, ssl_version=ssl.PROTOCOL_SSLv23,
                            ca_certs=self.ssl["ca_certs"])
                except IOError as exc:
                    self._socket.close()
                    raise ReqlDriverError("SSL handshake failed (see server log for more information): %s" % str(exc))
                try:
                    match_hostname(self._socket.getpeercert(), hostname=self.host)
                except CertificateError:
                    self._socket.close()
                    raise

            parent._parent.handshake.reset()
            response = None
            while True:
                request = parent._parent.handshake.next_message(response)
                if request is None:
                    break
                # This may happen in the `V1_0` protocol where we send two requests as
                # an optimization, then need to read each separately
                if request is not "":
                    self.sendall(request)

                # The response from the server is a null-terminated string
                response = b''
                while True:
                    char = self.recvall(1)
                    if char == b'\0':
                        break
                    response += char
        except (ReqlAuthError, ReqlTimeoutError):
            self.close()
            raise
        except ReqlDriverError as ex:
            self.close()
            error = str(ex)\
                .replace('receiving from', 'during handshake with')\
                .replace('sending to', 'during handshake with')
            raise ReqlDriverError(error)
        except Exception as ex:
            self.close()
            raise ReqlDriverError("Could not connect to %s:%s. Error: %s" %
                                  (self.host, self.port, ex))

    def close(self):
        if self._socket is not None:
            try:
                self._socket.shutdown(socket.SHUT_RDWR)
                self._socket.close()
            except Exception:
                pass
            finally:
                self._socket = None

    def recvall(self, length):
        res = b'' if self._read_buffer is None else self._read_buffer
        while len(res) < length:
            while True:
                try:
                    chunk = self._socket.recv(length - len(res))
                    break
                except ReqlTimeoutError:
                    raise
                except IOError as ex:
                    if ex.errno == errno.ECONNRESET:
                        self.close()
                        raise ReqlDriverError("Connection is closed.")
                    elif ex.errno != errno.EINTR:
                        self.close()
                        raise ReqlDriverError(('Connection interrupted ' +
                                              'receiving from %s:%s - %s') %
                                             (self.host, self.port, str(ex)))
                except Exception as ex:
                    self.close()
                    raise ReqlDriverError('Error receiving from %s:%s - %s' %
                                         (self.host, self.port, str(ex)))
                except:
                    self.close()
                    raise
            if len(chunk) == 0:
                self.close()
                raise ReqlDriverError("Connection is closed.")
            res += chunk
        return res

    def sendall(self, data):
        offset = 0
        while offset < len(data):
            try:
                offset += self._socket.send(data[offset:])
            except IOError as ex:
                if ex.errno == errno.ECONNRESET:
                    self.close()
                    raise ReqlDriverError("Connection is closed.")
                elif ex.errno != errno.EINTR:
                    self.close()
                    raise ReqlDriverError(('Connection interrupted ' +
                                          'sending to %s:%s - %s') %
                                         (self.host, self.port, str(ex)))
            except Exception as ex:
                self.close()
                raise ReqlDriverError('Error sending to %s:%s - %s' %
                                     (self.host, self.port, str(ex)))
            except:
                self.close()
                raise


class ConnectionInstance(object):
    def __init__(self, parent, io_loop=None):
        self._parent = parent
        self._closing = False
        self._user_queries = { }
        self._cursor_cache = { }

        self._write_mutex = Semaphore()
        self._socket = None

    def connect(self, timeout):
        with gevent.Timeout(timeout, RqlTimeoutError(self._parent.host, self._parent.port)) as timeout:
            self._socket = SocketWrapper(self)

        # Start a parallel coroutine to perform reads
        gevent.spawn(self._reader)
        return self._parent

    def is_open(self):
        return self._socket is not None and self._socket.is_open()

    def close(self, noreply_wait=False, token=None, exception=None):
        self._closing = True
        if exception is not None:
            err_message = "Connection is closed (%s)." % str(exception)
        else:
            err_message = "Connection is closed."

        # Cursors may remove themselves when errored, so copy a list of them
        for cursor in list(self._cursor_cache.values()):
            cursor._error(err_message)

        for query, async_res in iter(self._user_queries.values()):
            async_res.set_exception(RqlDriverError(err_message))

        self._user_queries = { }
        self._cursor_cache = { }

        if noreply_wait:
            noreply = net.Query(pQuery.NOREPLY_WAIT, token, None, None)
            self.run_query(noreply, False)

        try:
            self._socket.close()
        except:
            pass

    # TODO: make connection recoverable if interrupted by a user's gevent.Timeout?
    def run_query(self, query, noreply):
        self._write_mutex.acquire()

        try:
            self._socket.sendall(query.serialize(self._parent._get_json_encoder(query)))
        finally:
            self._write_mutex.release()

        if noreply:
            return None

        async_res = AsyncResult()
        self._user_queries[query.token] = (query, async_res)
        return async_res.get()

    # The _reader coroutine runs in its own coroutine in parallel, reading responses
    # off of the socket and forwarding them to the appropriate AsyncResult or Cursor.
    # This is shut down as a consequence of closing the stream, or an error in the
    # socket/protocol from the server.  Unexpected errors in this coroutine will
    # close the ConnectionInstance and be passed to any open AsyncResult or Cursors.
    def _reader(self):
        try:
            while True:
                buf = self._socket.recvall(12)
                (token, length,) = struct.unpack("<qL", buf)
                buf = self._socket.recvall(length)

                cursor = self._cursor_cache.get(token)
                if cursor is not None:
                    cursor._extend(buf)
                elif token in self._user_queries:
                    # Do not pop the query from the dict until later, so
                    # we don't lose track of it in case of an exception
                    query, async_res = self._user_queries[token]
                    res = net.Response(token, buf, self._parent._get_json_decoder(query))
                    if res.type == pResponse.SUCCESS_ATOM:
                        async_res.set(net.maybe_profile(res.data[0], res))
                    elif res.type in (pResponse.SUCCESS_SEQUENCE,
                                      pResponse.SUCCESS_PARTIAL):
                        cursor = GeventCursor(self, query, res)
                        async_res.set(net.maybe_profile(cursor, res))
                    elif res.type == pResponse.WAIT_COMPLETE:
                        async_res.set(None)
                    else:
                        async_res.set_exception(res.make_error(query))
                    del self._user_queries[token]
                elif not self._closing:
                    raise RqlDriverError("Unexpected response received.")
        except Exception as ex:
            if not self._closing:
                self.close(exception=ex)


class Connection(net.Connection):
    def __init__(self, *args, **kwargs):
        super(Connection, self).__init__(ConnectionInstance, *args, **kwargs)
